@@ -1,12 +1,13 @@
 from operator import methodcaller, attrgetter
-from abc import ABCMeta
+from abc import ABCMeta, abstractmethod
+from copy import deepcopy
 import re
 import os
 
 from ccp_util import IPv4Obj
 
 """ ccp_abc.py - Parse, Query, Build, and Modify IOS-style configurations
-     Copyright (C) 2014 David Michael Pennington
+     Copyright (C) 2014-2015 David Michael Pennington
 
      This program is free software: you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
@@ -52,7 +53,6 @@ class BaseCfgLine(object):
         self.feature   = ''        # Major feature description
         self.feature_param1 = ''   # Parameter1 of the feature
         self.feature_param2 = ''   # Parameter2 of the feature (if req'd)
-        self.hash_arg = None
 
         self.set_comment_bool()
 
@@ -63,49 +63,61 @@ class BaseCfgLine(object):
             return "<%s # %s '%s' (parent is # %s)>" % (self.classname, 
                 self.linenum, self.text, self.parent.linenum)
 
-
     def __str__(self):
         return self.__repr__()
 
-    def __eq__(self, val):
-        try:
-            ##   try / except is much faster than isinstance();
-            ##   I added hash_arg() inline below for speed... whenever I change
-            ##   hash_arg(), I *must* change this
-            return (str(self.linenum)+self.text)==(str(val.linenum)+val.text)
-        except:
-            return False
+    def __hash__(self):
+        ##   I inlined the hash() argument below for speed... whenever I change
+        ##   self.__eq__() I *must* change this
+        return hash(str(self.linenum)+self.text)
 
     def __gt__(self, val):
         if (self.linenum>val.linenum):
             return True
         return False
 
+    def __eq__(self, val):
+        try:
+            ##   try / except is much faster than isinstance();
+            ##   I added hash_arg() inline below for speed... whenever I change
+            ##   self.__hash__() I *must* change this
+            return (str(self.linenum)+self.text)==(str(val.linenum)+val.text)
+        except:
+            return False
+
     def __lt__(self, val):
         # Ref: http://stackoverflow.com/a/7152796/667301
         if (self.linenum<val.linenum):
             return True
         return False
 
-    def __hash__(self):
-        return hash(self.hash_arg)
-
-    def hash_arg(self):
-        # Just a unique string or each object instance
-        return str(self.linenum)+self.text
-
     def set_comment_bool(self):
+        delimiters = set(self.comment_delimiter)
         retval = None
         ## Use this instead of a regex... nontrivial speed enhancement
         tmp = self.text.lstrip()
-        if len(tmp)>0 and \
-            (self.comment_delimiter==tmp[len(self.comment_delimiter)-1]):
-            retval = True
-        else:
-            retval = False
+        for delimit_char in delimiters:
+            if len(tmp)>0 and \
+                (delimit_char==tmp[len(delimit_char)-1]):
+                retval = True
+                break
+            else:
+                retval = False
         self.is_comment = retval
         return retval
 
+    @property
+    def dna(self):
+        return self.classname
+
+    @property
+    def hash_children(self):
+        """Return a unique hash of all children (if the number of children > 0)"""
+        if len(self.children)>0:
+            return hash(tuple(self.children))
+        else:
+            return 0
+
     @property
     def family_endpoint(self):
         if self.children==[]:
@@ -148,6 +160,13 @@ class BaseCfgLine(object):
             return True
         return False
 
+    @property
+    def is_config_line(self):
+        """Return a boolean for whether this is a config statement; returns False if this object is a blank line, or a comment"""
+        if len(self.text.strip())>0 and not self.is_comment:
+            return True
+        return False
+
 
     def _list_reassign_linenums(self):
         # Call this when I want to reparse everything
@@ -188,8 +207,14 @@ class BaseCfgLine(object):
         if recurse:
             for child in self.children:
                 child.delete()
-        del self.confobj._list[self.linenum]
-        self._list_reassign_linenums()
+
+        ## Consistency check to refuse deletion of the wrong object...
+        ##    only delete if the line numbers are consistent
+        text = self.text
+        linenum = self.linenum
+        if self.confobj._list[self.linenum].text==text:
+            del self.confobj._list[self.linenum]
+            self._list_reassign_linenums()
 
     def delete_children_matching(self, linespec):
         """Delete any child :class:`~models_cisco.IOSCfgLine` objects which 
@@ -259,12 +284,16 @@ class BaseCfgLine(object):
         retval = self.confobj.insert_after(self, insertstr, atomic=local_atomic)
         return retval
 
-    def append_to_family(self, insertstr):
+    def append_to_family(self, insertstr, indent=-1, auto_indent_width=1, 
+        auto_indent=False):
         """Append an :class:`~models_cisco.IOSCfgLine` object with ``insertstr``
-        to the bottom of the current configuration family.
+        as a child at the bottom of the current configuration family.
 
         Args:
             - insertstr (str): A string which contains the text configuration to be apppended.
+            - indent (int): The amount of indentation to use for the child line; by default, the number of left spaces provided with ``insertstr`` are respected.  However, you can manually set the indent level when ``indent``>0.  This option will be ignored, if ``auto_indent`` is True.
+            - auto_indent_width (int): Amount of whitespace to automatically indent
+            - auto_indent (bool): Automatically indent the child to ``auto_indent_width``
 
         Returns:
             - str.  The text matched by the regular expression group; if there is no match, None is returned.
@@ -304,12 +333,24 @@ class BaseCfgLine(object):
            !
            >>>
         """
+        ## Build the string to insert with proper indentation...
+        if auto_indent:
+            insertstr = (" "*(self.indent+auto_indent_width))+insertstr.lstrip()
+        elif indent>0:
+            insertstr = (" "*(self.indent+indent))+insertstr.lstrip()
+
         ## BaseCfgLine.append_to_family(), insert a single line after this 
         ##  object's children
         local_atomic=False
-        last_child = self.children[-1]
-        retval = self.confobj.insert_after(last_child, insertstr, 
-            atomic=local_atomic)
+        try:
+            last_child = self.children[-1]
+            retval = self.confobj.insert_after(last_child, insertstr, 
+                atomic=local_atomic)
+        except IndexError:
+            # The object has no children
+            retval = self.confobj.insert_after(self, insertstr, 
+                atomic=local_atomic)
+
         return retval
 
     def replace(self, linespec, replacestr, ignore_rgx=None):
@@ -422,7 +463,7 @@ class BaseCfgLine(object):
         Args:
             - regex (str): A string or python regular expression, which should be matched.  This regular expression should contain parenthesis, which bound a match group.
         Kwargs:
-            - group (int): An integer which specifies the desired group to be returned.  ``group`` defaults to 1.
+            - group (int): An integer which specifies the desired regex group to be returned.  ``group`` defaults to 1.
             - default (str): The default value to be returned, if there is no match.  By default an empty string is returned if there is no match.
 
         Returns:
@@ -488,22 +529,18 @@ class BaseCfgLine(object):
             - list.  A list of matching :class:`~models_cisco.IOSCfgLine` objects which matched.  If there is no match, an empty :py:func:`list` is returned.
 
         """
-        retval = list()
-        for cobj in self.children:
-            if cobj.re_search(regex):
-                retval.append(cobj)
-        return retval
+        return [cobj for cobj in self.children if cobj.re_search(regex)]
 
     def re_match_typed(self, regex, group=1, result_type=str, default=''):
         """Use ``regex`` to search the :class:`~models_cisco.IOSCfgLine` text 
         and return the contents of the regular expression group, at the 
-        integer index, cast as ``result_type``; if there is no match, 
+        integer ``group`` index, cast as ``result_type``; if there is no match, 
         ``default`` is returned.
 
         Args:
             - regex (str): A string or python regular expression, which should be matched.  This regular expression should contain parenthesis, which bound a match group.
         Kwargs:
-            - group (int): An integer which specifies the desired group to be returned.  ``group`` defaults to 1.
+            - group (int): An integer which specifies the desired regex group to be returned.  ``group`` defaults to 1.
             - result_type (type): A type (typically one of: ``str``, ``int``, ``float``, or ``IPv4Obj``).  All returned values are cast as ``result_type``, which defaults to ``str``.
             - default (any): The default value to be returned, if there is no match.
 
@@ -552,21 +589,24 @@ class BaseCfgLine(object):
                 return result_type(default)
         return result_type(default)
 
-    def re_match_iter_typed(self, regex, group=1, result_type=str, default=''):
+    def re_match_iter_typed(self, regex, group=1, result_type=str, default='', 
+        all_children=False):
         """Use ``regex`` to search the children of 
         :class:`~models_cisco.IOSCfgLine` text and return the contents of 
-        the regular expression group, at the integer index, cast as 
+        the regular expression group, at the integer ``group`` index, cast as 
         ``result_type``; if there is no match, ``default`` is returned.
 
         Args:
             - regex (str): A string or python compiled regular expression, which should be matched.  This regular expression should contain parenthesis, which bound a match group.
         Kwargs:
-            - group (int): An integer which specifies the desired group to be returned.  ``group`` defaults to 1.
+            - group (int): An integer which specifies the desired regex group to be returned.  ``group`` defaults to 1.
             - result_type (type): A type (typically one of: ``str``, ``int``, ``float``, or :class:`~ccp_util.IPv4Obj`).         All returned values are cast as ``result_type``, which defaults to ``str``.
             - default (any): The default value to be returned, if there is no match.
+            - all_children (bool): Set True if you want to search all children (children, grand children, great grand children, etc...)
 
         Returns:
             - ``result_type``.  The text matched by the regular expression group; if there is no match, ``default`` is returned.  All values are cast as ``result_type``.
+        - NOTE: This loops through the children (in order) and returns when the regex hits its first match.
 
         This example illustrates how you can use 
         :func:`~models_cisco.IOSCfgLine.re_match_iter_typed` to build an 
@@ -603,11 +643,18 @@ class BaseCfgLine(object):
             ##   this while I build the API
             raise NotImplementedError
 
-        for cobj in self.children:
-            mm = re.search(regex, cobj.text)
-            if not (mm is None):
-                return result_type(mm.group(group))
-        return result_type(default)
+        if all_children is False:
+            for cobj in self.children:
+                mm = re.search(regex, cobj.text)
+                if not (mm is None):
+                    return result_type(mm.group(group))
+            return result_type(default)
+        else:
+            for cobj in self.all_children:
+                mm = re.search(regex, cobj.text)
+                if not (mm is None):
+                    return result_type(mm.group(group))
+            return result_type(default)
 
     def reset(self):
         # For subclass APIs
@@ -617,6 +664,14 @@ class BaseCfgLine(object):
         # For subclass APIs
         raise NotImplementedError
 
+    @property
+    def ioscfg(self):
+        """Return a list with this the text of this object, and 
+        with all children in the direct line."""
+        retval = [self.text]
+        retval.extend(list(map(attrgetter('text'), self.all_children)))
+        return retval
+
     @property
     def lineage(self):
         """Iterate through to the oldest ancestor of this object, and return
@@ -629,6 +684,26 @@ class BaseCfgLine(object):
             retval.extend(self.all_children)
         return sorted(retval)
 
+    @property
+    def geneology(self):
+        """Iterate through to the oldest ancestor of this object, and return
+        a list of all ancestors in the direct line as well as this obj.  
+        Cousins or aunts / uncles are *not* returned.  Note: children of this 
+        object are *not* returned."""
+        retval = sorted(self.all_parents)
+        retval.append(self)
+        return retval
+
+    @property
+    def geneology_text(self):
+        """Iterate through to the oldest ancestor of this object, and return
+        a list of all ancestors in the direct line as well as this obj.  
+        Cousins or aunts / uncles are *not* returned.  Note: children of this 
+        object are *not* returned."""
+        retval = map(lambda x: x.text, sorted(self.all_parents))
+        retval.append(self.text)
+        return retval
+
     @property
     def is_parent(self):
         return bool(self.has_children)
