@@ -1,8 +1,6 @@
-import sys
 import re
-import os
 
-from protocol_values import ASA_TCP_PORTS, ASA_UDP_PORTS
+from protocol_values import ASA_TCP_PORTS, ASA_UDP_PORTS, ASA_IP_PROTOCOLS
 from ccp_abc import BaseCfgLine
 from ccp_util import L4Object
 from ccp_util import IPv4Obj
@@ -16,7 +14,7 @@ from ccp_util import IPv4Obj
 ###   Use models_asa.py at your own risk.  You have been warned :-)
 
 """ models_asa.py - Parse, Query, Build, and Modify IOS-style configurations
-     Copyright (C) 2014 David Michael Pennington
+     Copyright (C) 2014-2015 David Michael Pennington
 
      This program is free software: you can redistribute it and/or modify
      it under the terms of the GNU General Public License as published by
@@ -227,6 +225,15 @@ class BaseASAIntfLine(ASACfgLine):
         except:
             return self.default_ipv4_addr_object
 
+    @property
+    def ipv4_standby_addr_object(self):
+        """Return a ccp_util.IPv4Obj object representing the standby address on this interface; if there is no address, return IPv4Obj('127.0.0.1/32')"""
+        try:
+            return IPv4Obj('%s/%s' % (self.ipv4_standby_addr, 
+                self.ipv4_netmask))
+        except:
+            return self.default_ipv4_addr_object
+
     @property
     def ipv4_network_object(self):
         """Return an ccp_util.IPv4Obj object representing the subnet on this interface; if there is no address, return ccp_util.IPv4Obj('127.0.0.1/32')"""
@@ -280,14 +287,21 @@ class BaseASAIntfLine(ASACfgLine):
     @property
     def ipv4_addr(self):
         """Return a string with the interface's IPv4 address, or '' if there is none"""
-        retval = self.re_match_iter_typed(r'^\s+ip\s+address\s+(\d+\.\d+\.\d+\.\d+)\s+\d+\.\d+\.\d+\.\d+\s*$', 
+        retval = self.re_match_iter_typed(r'^\s+ip\s+address\s+(\d+\.\d+\.\d+\.\d+)\s+\d+\.\d+\.\d+\.\d+(\sstandby\s+\S+\s*)*$', 
+            result_type=str, default='')
+        return retval
+
+    @property
+    def ipv4_standby_addr(self):
+        """Return a string with the interface's IPv4 address, or '' if there is none"""
+        retval = self.re_match_iter_typed(r'^\s+ip\s+address\s+\d+\.\d+\.\d+\.\d+\s+\d+\.\d+\.\d+\.\d+\sstandby\s+(\S+)\s*$', 
             result_type=str, default='')
         return retval
 
     @property
     def ipv4_netmask(self):
         """Return a string with the interface's IPv4 netmask, or '' if there is none"""
-        retval = self.re_match_iter_typed(r'^\s+ip\s+address\s+\d+\.\d+\.\d+\.\d+\s+(\d+\.\d+\.\d+\.\d+)\s*$',
+        retval = self.re_match_iter_typed(r'^\s+ip\s+address\s+\d+\.\d+\.\d+\.\d+\s+(\d+\.\d+\.\d+\.\d+)(\sstandby\s+\S+\s*)*$',
             result_type=str, default='')
         return retval
 
@@ -367,20 +381,39 @@ class BaseASAIntfLine(ASACfgLine):
 ##-------------  ASA name
 ##
 
+_RE_NAMEOBJECT_STR = r'^name\s+(?P<addr>\d+\.\d+\.\d+\.\d+)\s(?P<name>\S+)'
+_RE_NAMEOBJECT = re.compile(_RE_NAMEOBJECT_STR, re.VERBOSE)
 class ASAName(ASACfgLine):
 
     def __init__(self, *args, **kwargs):
         """Accept an ASA line number and initialize family relationship
         attributes"""
         super(ASAName, self).__init__(*args, **kwargs)
+        mm = _RE_NAMEOBJECT.search(self.text)
+        if not (mm is None):
+            self._mm_results = mm.groupdict()   # All regex match results
+        else:
+            raise ValueError
+
+        self.name = self._mm_results['name']
+        self.addr = self._mm_results['addr']
 
     @classmethod
     def is_object_for(cls, line="", re=re):
-        name_regex = r'^name\s+(\d+\.\d+\.\d+\.\d+)\s(\S+)'
-        if re.search(name_regex, line):
+        if 'name ' in line[0:5].lower():
             return True
         return False
 
+    @property
+    def result_dict(self):
+        mm_r = self._mm_results
+        retval = dict()
+
+        retval['name'] = self._mm_results['name']
+        retval['addr'] = self._mm_results['addr']
+
+        return retval
+
 ##
 ##-------------  ASA object network
 ##
@@ -394,8 +427,7 @@ class ASAObjNetwork(ASACfgLine):
 
     @classmethod
     def is_object_for(cls, line="", re=re):
-        obj_group_regex = r'^object\s+network\s+(\S+)'
-        if re.search(obj_group_regex, line):
+        if 'object network ' in line[0:15].lower():
             return True
         return False
 
@@ -412,18 +444,20 @@ class ASAObjService(ASACfgLine):
 
     @classmethod
     def is_object_for(cls, line="", re=re):
-        obj_group_regex = r'^object\s+service\s+(\S+)'
-        if re.search(obj_group_regex, line):
+        if 'object service ' in line[0:15].lower():
             return True
         return False
 
 ##
 ##-------------  ASA object-group network
 ##
-
-_RE_NETHOST = re.compile(r'^\s*network-object\s+host\s+(\S+)')
-_RE_NETWORK1 = re.compile(r'^\s*network-object\s+(\S+)')
-_RE_NETWORK2 = re.compile(r'^\s*network-object\s+\S+\s+(\d+\.\d+\.\d+\.\d+)')
+_RE_NETOBJECT_STR = r"""(?:                         # Non-capturing parenthesis
+ (^\s*network-object\s+host\s+(?P<host>\S+))
+|(^\s*network-object\s+(?P<network>\S+)\s+(?P<netmask>\d+\.\d+\.\d+\.\d+))
+|(^\s*group-object\s+(?P<groupobject>\S+))
+)                                                   # Close non-capture parens
+"""
+_RE_NETOBJECT = re.compile(_RE_NETOBJECT_STR, re.VERBOSE)
 class ASAObjGroupNetwork(ASACfgLine):
 
     def __init__(self, *args, **kwargs):
@@ -436,46 +470,95 @@ class ASAObjGroupNetwork(ASACfgLine):
 
     @classmethod
     def is_object_for(cls, line="", re=re):
-        obj_group_regex = r'^object-group\s+network\s+(\S+)'
-        if re.search(obj_group_regex, line):
+        if 'object-group network ' in line[0:21].lower():
             return True
         return False
 
     @property
-    def networks(self):
-        """Return a list of IPv4Obj objects which represent the address space allowed by
-        This object-group"""
+    def hash_children(self):
+        ## Manually override the BaseCfgLine method since this recurses through
+        ##    children
+        ## FIXME: Implement hash_children for ASAObjGroupService
+        return hash(tuple(self.network_strings))  # network_strings recurses...
+
+    @property
+    def network_count(self):
+        ## Return the number of discrete network objects covered by this group
+        ## FIXME: Implement port_count for ASAObjGroupService
+        return len(self.network_strings)
+
+    @property
+    def network_strings(self):
+        """Return a list of strings which represent the address space allowed by
+        this object-group"""
         retval = list()
         names = self.confobj.names
         for obj in self.children:
-            if 'network-object host ' in obj.text:
-                host_str = obj.re_match_typed(_RE_NETHOST, group=1, result_type=str)
-                retval.append(IPv4Obj(names.get(host_str, host_str)))
-            elif 'network-object ' in obj.text:
-                network_str = obj.re_match_typed(_RE_NETWORK1, group=1, result_type=str)
-                netmask_str = obj.re_match_typed(_RE_NETWORK2, group=1, result_type=str)
-                retval.append(IPv4Obj('{0} {1}'.format(names.get(network_str, 
-                    network_str), netmask_str)))
-            elif 'group-object ' in obj.text:
-                name = obj.re_match_typed(r'^\s*group-object\s+(\S+)', group=1, 
-                    result_type=str)
-                group_nets = self.confobj.object_group_network.get(name, None)
-                if name==self.name:
+
+            ## Parse out 'object-group ...' and 'group-object' lines...
+            mm = _RE_NETOBJECT.search(obj.text)
+            if not (mm is None):
+                net_obj = mm.groupdict()
+                if net_obj['netmask']=='255.255.255.255':
+                    net_obj['host'] = net_obj['network']
+            else:
+                net_obj = dict()
+
+            if net_obj.get('host', None):
+                retval.append(names.get(net_obj['host'], 
+                    net_obj['host']))
+            elif net_obj.get('network', None):
+                ## This is a non-host network object
+                retval.append('{0}/{1}'.format(names.get(net_obj['network'], 
+                    net_obj['network']), net_obj['netmask']))
+            elif net_obj.get('groupobject', None):
+                groupobject = net_obj['groupobject']
+                if groupobject==self.name:
                     ## Throw an error when importing self
-                    raise ValueError("FATAL: Cannot recurse through group-object {0} in object-group network {1}".format(name, self.name))
+                    raise ValueError("FATAL: Cannot recurse through group-object {0} in object-group network {1}".format(groupobject, self.name))
+
+                group_nets = self.confobj.object_group_network.get(groupobject,
+                    None)
                 if (group_nets is None):
                     raise ValueError("FATAL: Cannot find group-object named {0}".format(name))
                 else:
-                    retval.extend(group_nets.networks)
+                    retval.extend(group_nets.network_strings)
             elif 'description ' in obj.text:
                 pass
             else:
                 raise NotImplementedError("Cannot parse '{0}'".format(obj.text))
         return retval
 
+    @property
+    def networks(self):
+        """Return a list of IPv4Obj objects which represent the address space allowed by
+        This object-group"""
+        ## FIXME: Implement object caching for other ASAConfigList objects
+        ## Return a cached result if the networks lookup has already been done
+
+        retval = list()
+        for net_str in self.network_strings:
+            ## Check the ASACfgList cache of network objects
+            if not self.confobj._network_cache.get(net_str, False):
+                net = IPv4Obj(net_str)
+                self.confobj._network_cache[net_str] = net
+                retval.append(net)
+            else:
+                retval.append(self.confobj._network_cache[net_str])
+
+        return retval
+
 ##
 ##-------------  ASA object-group service
 ##
+_RE_PORTOBJ_STR = r"""(?:                            # Non-capturing parentesis
+ # service-object udp destination eq dns
+ (^\s*service-object\s+(?P<protocol>{0})\s+(?P<src_dst>\S+)\s+(?P<s_port>\S+))
+|(^\s*port-object\s+(?P<operator>eq|range)\s+(?P<p_port>\S.+))
+|(^\s*group-object\s+(?P<groupobject>\S+))
+)                                                   # Close non-capture parens
+""".format('tcp|udp|tcp-udp')
+_RE_PORTOBJECT = re.compile(_RE_PORTOBJ_STR, re.VERBOSE)
 
 class ASAObjGroupService(ASACfgLine):
 
@@ -499,8 +582,7 @@ class ASAObjGroupService(ASACfgLine):
 
     @classmethod
     def is_object_for(cls, line="", re=re):
-        obj_group_regex = r'^object-group\s+service\s+(\S+)'
-        if re.search(obj_group_regex, line):
+        if 'object-group service ' in line[0:21].lower():
             return True
         return False
 
@@ -514,16 +596,19 @@ class ASAObjGroupService(ASACfgLine):
         ## TODO: implement processing for group-objects (which obviously 
         ##    involves iteration
         #GROUP_OBJ_REGEX = r'^\s*group-object\s+(\S+)'
-        SERVICE_OBJ_REGEX = r'^\s*service-object\s+(tcp|udp|tcp-udp)\s+(\S+)\s+(\S+)'
-        PORT_OBJ_REGEX = r'^\s*port-object\s+(eq|range)\s+(\S.+)'
         for obj in self.children:
-            if 'service-object ' in obj.text:
-                protocol = obj.re_match_typed(SERVICE_OBJ_REGEX, 
-                    group=1, result_type=str)
-                src_dst = obj.re_match_typed(SERVICE_OBJ_REGEX, 
-                    group=2, result_type=str)
-                port = obj.re_match_typed(SERVICE_OBJ_REGEX, 
-                    group=3, result_type=str)
+
+            ## Parse out 'service-object ...' and 'port-object' lines...
+            mm = _RE_PORTOBJECT.search(obj.text)
+            if not (mm is None):
+                svc_obj = mm.groupdict()
+            else:
+                svc_obj = dict()
+
+            if svc_obj.get('protocol', None):
+                protocol = svc_obj.get('protocol')
+                src_dst = svc_obj.get('src_dst', '')
+                port = svc_obj.get('s_port', '')
 
                 if protocol=='tcp-udp':
                     retval.append(L4Object(protocol='tcp', 
@@ -534,13 +619,11 @@ class ASAObjGroupService(ASACfgLine):
                     retval.append(L4Object(protocol=protocol, 
                         port_spec=port, syntax='asa'))
 
-            elif 'port-object ' in obj.text:
-                op = obj.re_match_typed(PORT_OBJ_REGEX, 
-                    group=1, result_type=str)
-                port = obj.re_match_typed(PORT_OBJ_REGEX, 
-                    group=2, result_type=str)
-
+            elif svc_obj.get('operator', None):
+                op = svc_obj.get('operator', '')
+                port = svc_obj.get('p_port', '')
                 port_spec="{0} {1}".format(op, port)
+
                 if self.protocol_type=='tcp-udp':
                     retval.append(L4Object(protocol='tcp', 
                         port_spec=port_spec, syntax='asa'))
@@ -549,9 +632,9 @@ class ASAObjGroupService(ASACfgLine):
                 else:
                     retval.append(L4Object(protocol=self.protocol_type, 
                         port_spec=port_spec, syntax='asa'))
-            elif 'group-object ' in obj.text:
-                name = obj.re_match_typed(r'^\s*group-object\s+(\S+)', group=1, 
-                    result_type=str)
+
+            elif svc_obj.get('groupobject', None):
+                name = svc_obj.get('groupobject')
                 group_ports = self.confobj.object_group_service.get(name, None)
                 if name==self.name:
                     ## Throw an error when importing self
@@ -757,6 +840,266 @@ class ASARouteLine(BaseASARouteLine):
             group=2, result_type=str, default='')
         return retval
 
+_ACL_PROTOCOLS = 'ip|tcp|udp|ah|eigrp|esp|gre|igmp|igrp|ipinip|ipsec|ospf|pcp|pim|pptp|snp|\d+'
+_ACL_ICMP_PROTOCOLS = 'alternate-address|conversion-error|echo-reply|echo|information-reply|information-request|mask-reply|mask-request|mobile-redirect|parameter-problem|redirect|router-advertisement|router-solicitation|source-quench|time-exceeded|timestamp-reply|timestamp-request|traceroute|unreachable'
+_ACL_LOGLEVELS = r'alerts|critical|debugging|emergencies|errors|informational|notifications|warnings|[0-7]'
+_RE_ACLOBJECT_STR = r"""(?:                         # Non-capturing parenthesis
+# remark
+ (^access-list\s+(?P<acl_name0>\S+)\s+(?P<action0>remark)\s+(?P<remark>\S.+?)$)
+
+# extended service object with source network object, destination network object
+|(?:^access-list\s+(?P<acl_name1>\S+)
+  \s+extended\s+(?P<action1>permit|deny)
+  \s+(?:
+     (?:object-group\s+(?P<service_object1>\S+))
+    |(?P<protocol1>{0})
+  )
+  \s+(?:                       # 10.0.0.0 255.255.255.0
+     (?:object-group\s+(?P<src_networkobject1>\S+))
+    |(?:object\s+(?P<src_object1>\S+))
+    |(?:(?P<src_network1a>\S+)\s+(?P<src_netmask1a>\d+\.\d+\.\d+\.\d+))
+  )
+  \s+(?:                       # 10.0.0.0 255.255.255.0
+     (?:object-group\s+(?P<dst_networkobject1>\S+))
+    |(?:object\s+(?P<dst_object1>\S+))
+    |(?:(?P<dst_network1a>\S+)\s+(?P<dst_netmask1a>\d+\.\d+\.\d+\.\d+))
+  )
+  (?:\s+
+    (?P<log1>log)
+    (?:\s+(?P<loglevel1>{1}))?
+    (?:\s+interval\s+(?P<log_interval1>\d+))?
+  )?
+  (?:\s+(?P<disable1>disable))?
+  (?:
+    (?:\s+(?P<inactive1>inactive))
+   |(?:\s+time-range\s+(?P<time_range1>\S+))
+  )?
+ \s*$)    # END access-list 1 parse
+
+# extended service object with source network, destination network
+# access-list TESTME1 extended permit ip any any log
+|(?:^access-list\s+(?P<acl_name2>\S+)
+  \s+extended
+  \s+(?P<action2>permit|deny)
+  \s+(?:                       # service-object or protocol
+     (?:object-group\s+(?P<service_object2>\S+))
+    |(?P<protocol2>{0})
+  )
+  (?:\s+       # any, any4, host foo, object-group FOO or 10.0.0.0 255.255.255.0
+     (?:
+       (?P<src_network2a>any|any4)
+      |(?:host\s+(?P<src_network2b>\S+))
+      |(?:object\s+(?P<src_object2>\S+))
+      |(?:object-group\s+(?P<src_networkobject2>\S+))
+      |(?:(?P<src_network2c>\S+)\s+(?P<src_netmask2c>\d+\.\d+\.\d+\.\d+))
+    )
+  )
+  (?:\s+
+    (?:
+       (?:(?P<src_port_operator>eq|neq|lt|gt)\s+(?P<src_port>\S+))
+      |(?:range\s+(?P<src_port_low>\S+)\s+(?P<src_port_high>\S+))
+      |(?:object-group\s+(?P<src_service_object>\S+))
+    )
+  )?
+  (?:\s+       # any, any4, host foo, object-group FOO or 10.0.0.0 255.255.255.0
+     (?:
+       (?P<dst_network2a>any|any4)
+      |(?:host\s+(?P<dst_network2b>\S+))
+      |(?:object\s+(?P<dst_object2>\S+))
+      |(?:object-group\s+(?P<dst_networkobject2>\S+))
+      |(?:(?P<dst_network2c>\S+)\s+(?P<dst_netmask2c>\d+\.\d+\.\d+\.\d+))
+    )
+  )
+  (?:\s+
+    (?:
+       (?:(?P<dst_port_operator>eq|neq|lt|gt)\s+(?P<dst_port>\S+))
+      |(?:range\s+(?P<dst_port_low>\S+)\s+(?P<dst_port_high>\S+))
+      |(?:object-group\s+(?P<dst_service_object>\S+))
+    )
+  )?
+  (?:\s+
+    (?P<log2>log)
+    (?:\s+(?P<loglevel2>{1}))?
+    (?:\s+interval\s+(?P<log_interval2>\d+))?
+  )?
+  (?:\s+(?P<disable2>disable))?
+  (?:
+    (?:\s+(?P<inactive2>inactive))
+   |(?:\s+time-range\s+(?P<time_range2>\S+))
+  )?
+ \s*$)    # END access-list 2 parse
+
+# access-list SPLIT_TUNNEL_NETS standard permit 192.0.2.0 255.255.255.0
+|(?:^access-list\s+(?P<acl_name3>\S+)
+  \s+standard
+  \s+(?P<action3>permit|deny)
+  \s+(?:
+    (?P<dst_network3a>any|any4)
+   |(?:host\s+(?P<dst_network3b>\S+))
+   |(?:(?P<dst_network3c>\S+)\s+(?P<dst_netmask3c>\d+\.\d+\.\d+\.\d+))
+  )
+
+  )
+#access-list TESTME extended permit icmp any4 0.0.0.0 0.0.0.0 unreachable log interval 1
+|(?:^access-list\s+(?P<acl_name4>\S+)
+  \s+extended
+  \s+(?P<action4>permit|deny)
+  \s+(?P<protocol4>icmp)
+  (?:\s+       # any, any4, host foo, object-group FOO or 10.0.0.0 255.255.255.0
+     (?:
+       (?P<src_network4a>any|any4)
+      |(?:host\s+(?P<src_network4b>\S+))
+      |(?:object\s+(?P<src_object4>\S+))
+      |(?:object-group\s+(?P<src_networkobject4>\S+))
+      |(?:(?P<src_network4c>\S+)\s+(?P<src_netmask4c>\d+\.\d+\.\d+\.\d+))
+    )
+  )
+  (?:\s+       # any, any4, host foo, object-group FOO or 10.0.0.0 255.255.255.0
+     (?:
+       (?P<dst_network4a>any|any4)
+      |(?:host\s+(?P<dst_network4b>\S+))
+      |(?:object\s+(?P<dst_object4>\S+))
+      |(?:object-group\s+(?P<dst_networkobject4>\S+))
+      |(?:(?P<dst_network4c>\S+)\s+(?P<dst_netmask4c>\d+\.\d+\.\d+\.\d+))
+    )
+  )
+  (?:\s+(?P<icmp_proto4>{2}|\d+))?
+  (?:\s+
+    (?P<log4>log)
+    (?:\s+(?P<loglevel4>{1}))?
+    (?:\s+interval\s+(?P<log_interval4>\d+))?
+  )?
+  (?:\s+(?P<disable4>disable))?
+  (?:
+    (?:\s+(?P<inactive4>inactive))
+   |(?:\s+time-range\s+(?P<time_range4>\S+))
+  )?
+  )
+)                                                   # Close non-capture parens
+""".format(_ACL_PROTOCOLS, _ACL_LOGLEVELS, _ACL_ICMP_PROTOCOLS)
+_RE_ACLOBJECT = re.compile(_RE_ACLOBJECT_STR, re.VERBOSE)
+
+class ASAAclLine(ASACfgLine):
+
+    def __init__(self, *args, **kwargs):
+        """Provide attributes on Cisco ASA Access-Lists"""
+        super(ASAAclLine, self).__init__(*args, **kwargs)
+        mm = _RE_ACLOBJECT.search(self.text)
+        if not (mm is None):
+            self._mm_results = mm.groupdict()   # All regex match results
+        else:
+            raise ValueError("[FATAL] models_asa cannot parse '{0}'".format(self.text))
+
+    @classmethod
+    def is_object_for(cls, line="", re=re):
+        #if _RE_ACLOBJECT.search(line):
+        if 'access-list ' in line[0:13].lower():
+            return True
+        return False
+
+    @property
+    def src_addr_method(self):
+        mm_r = self._mm_results
+        if mm_r['action0'] and (mm_r['action0']=='remark'):
+            # remarks return an empty string
+            return ''
+        elif mm_r['src_networkobject1'] or mm_r['src_networkobject2'] or mm_r['src_networkobject4']:
+            return 'object-group'
+        elif mm_r['src_object1'] or mm_r['src_object2'] or mm_r['src_object4']:
+            return 'object'
+        elif mm_r['src_network1a'] or mm_r['src_network2a'] \
+            or mm_r['src_network2b'] or mm_r['src_network2c'] \
+            or mm_r['src_network4a'] or mm_r['src_network4b'] \
+            or mm_r['src_network4c']:
+            return 'network'
+        ## NOTE: I intended to match dst addrs here...
+        elif mm_r['acl_name3']:
+            ## Special case: standard ACLs match any src implicitly
+            self._mm_results['src_network3'] = 'any4'
+            return 'network'
+        else:
+            raise ValueError("Cannot parse ACL source address method for '{0}'".format(self.text))
+
+    @property
+    def dst_addr_method(self):
+        mm_r = self._mm_results
+        if mm_r['action0'] and (mm_r['action0']=='remark'):
+            # remarks return an empty string
+            return ''
+        elif mm_r['dst_networkobject1'] or mm_r['dst_networkobject2'] or mm_r['dst_networkobject4']:
+            return 'object-group'
+        elif mm_r['dst_object1'] or mm_r['dst_object2'] or mm_r['dst_object4']:
+            return 'object'
+        elif mm_r['dst_network1a'] or mm_r['dst_network2a'] \
+            or mm_r['dst_network2b'] or mm_r['dst_network2c'] \
+            or mm_r['dst_network4a'] or mm_r['dst_network4b'] \
+            or mm_r['dst_network4c']:
+            return 'network'
+        elif mm_r['dst_network3a'] or mm_r['dst_network3b'] \
+            or mm_r['dst_network3c']:
+            return 'network'
+        else:
+            raise ValueError("Cannot parse ACL destination address method for '{0}'".format(self.text))
+
+    @property
+    def acl_protocol_dict(self):
+        mm_r = self._mm_results
+        retval = dict()
+
+        if mm_r['remark']:
+            # remarks get IP protocol -1
+            retval['protocol'] = -1
+            retval['protocol_object'] = ''
+            return retval
+        elif mm_r['protocol1'] or mm_r['protocol2'] or mm_r['protocol4']:
+            _proto = mm_r['protocol1'] or mm_r['protocol2'] or mm_r['protocol4'] or -1
+            retval['protocol'] = int(ASA_IP_PROTOCOLS.get(_proto, _proto))
+            retval['protocol_object'] = ''
+            return retval
+        elif mm_r['acl_name3']:
+            # Special case for standard ASA ACLs
+            _proto = 'ip'
+            retval['protocol'] = int(ASA_IP_PROTOCOLS.get(_proto, _proto))
+            retval['protocol_object'] = ''
+            return retval
+        elif mm_r['service_object1'] or mm_r['service_object2']:
+            # protocol service objects get a special protocol number
+            retval['protocol'] = 65535
+            retval['protocol_object'] = mm_r['service_object1'] \
+                or mm_r['service_object2']
+            return retval
+        else:
+            raise ValueError("Cannot parse ACL protocol value for '{0}'".format(self.text))
+
+    @property
+    def result_dict(self):
+        mm_r = self._mm_results
+        retval = dict()
+
+        proto_dict = self.acl_protocol_dict
+        retval['ip_protocol'] = proto_dict['protocol']
+        retval['ip_protocol_object'] = proto_dict['protocol_object']
+        retval['acl_name'] = mm_r['acl_name0'] or mm_r['acl_name1'] \
+            or mm_r['acl_name2'] or mm_r['acl_name3'] or mm_r['acl_name4']
+        retval['action'] = mm_r['action0'] or mm_r['action1'] \
+            or mm_r['action2'] or mm_r['action3'] or mm_r['action4']
+        retval['remark'] = mm_r['remark']
+        retval['src_addr_method'] = self.src_addr_method
+        retval['dst_addr_method'] = self.dst_addr_method
+        retval['disable'] = bool(mm_r['disable1'] or mm_r['disable2'] or mm_r['disable4'])
+        retval['time_range'] = mm_r['time_range1'] or mm_r['time_range2'] or mm_r['time_range4']
+        retval['log'] = bool(mm_r['log1'] or mm_r['log2'] or mm_r['log4'])
+        if not retval['log']:
+            retval['log_interval'] = -1
+            retval['log_level'] = ''
+        else:
+            retval['log_level'] = mm_r['loglevel1'] or mm_r['loglevel2'] or mm_r['loglevel4'] or 'informational'
+            retval['log_interval'] = int(mm_r['log_interval1'] \
+                or mm_r['log_interval2'] or mm_r['log_interval4'] or 300)
+
+        return retval
+
+
 ################################
 ################################ Groups ###############################
 ################################
